import numpy as np

import tensorflow as tf
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import constraints, initializers, regularizers

from quple.interface.tfq.layers import PQC
from quple.interface.tfq.tf_utils import get_output_shape

class QConv2D(PQC):
    """Quantum Convolution 2D Layer.
    Quantum convolution uses a quantum filter as the basic building block. 
    It replaces the classical filter by a Parameterised Quantum Circuit (PQC) 
    which scan across the local regions (the receptive field) of an image. 
    A quantum filter transforms the pixel values in the local region into the 
    quantum states of its data circuit via a suitable feature map, hence projecting
    the data into a higher dimensional quantum (Hilbert) space. The quantum states 
    are then propagated to the kernel circuit of the quantum filter which undergoes a 
    sequence parameterised gate operations and outputs the expectation value of a 
    given measurement operator.
    """
    def __init__(self, kernel_circuit,
                 data_circuit,
                 operators,
                 kernel_size,
                 filters=1,
                 strides=1,
                 padding='same',
                 initializer=None,
                 regularizer=None,
                 constraint=None,
                 parameter_sharing=True,
                 trainable=True,       
                 seed=None,
                 name=None, **kwargs):
        """Instantiate this layer.
        
        Arguments:
            kernel_circuit: `cirq.Circuit` or `quple.QuantumCircuit` containing 
                `sympy.Symbols` that will be used as the model which will be fed with 
                quantum data from the data circuit.
            data_circuit: `cirq.Circuit` or `quple.QuantumCircuit` containing 
                `sympy.Symbols` that maps the input patches in this layer to the quantum
                data of the circuit which will be fed to the model circuit.
            operators: `cirq.PauliSum` or Python `list` of `cirq.PauliSum` objects
                used as observables at the end of the kernel circuit.
            filters: Integer, the dimensionality of the output space (i.e. the number
              of filters in the convolution).
            kernel_size: An integer or tuple/list of n integers, specifying the
              length of the convolution window.
            strides: An integer or tuple/list of n integers,
              specifying the stride length of the convolution.
              Specifying any stride value != 1 is incompatible with specifying
              any `dilation_rate` value != 1.
            padding: One of `"valid"`,  `"same"`, or `"causal"` (case-insensitive).
              `"valid"` means no padding. `"same"` results in padding with zeros
              evenly to the left/right or up/down of the input such that output has the
              same height/width dimension as the input. `"causal"` results in causal
              (dilated) convolutions, e.g. `output[t]` does not depend on `input[t+1:]`.
            initializer: Optional `tf.keras.initializer` object to specify how the
                symbols in `kernel_circuit` should be initialized when creating
                the managed variables. If None, defaults to initialize from a
                random uniform distribution between 0 and 2*pi. 
            regularizer: Optional `tf.keras.regularizer` object applied to the
                managed variables parameterizing `kernel_circuit`.
            constraint: Optional `tf.keras.constraint` object applied to the
                managed variables parameterizing `kernel_circuit`.
            parameter_sharing: Boolean. Whether to share parameters for quantum filters
            across different local regions of the input data.
            trainable: Boolean, whether the layer's variables should be trainable.
            seed: Optional integer. Controls the random seed of the weight initializer.
            name: Optional string. Name of the layer.
        """
        self.rank = 2
        self.filters = filters
        self.kernel_size = conv_utils.normalize_tuple(kernel_size, self.rank, 'kernel_size')
        self.strides = conv_utils.normalize_tuple(strides, self.rank, 'strides')
        self.padding = conv_utils.normalize_padding(padding)
        self.parameter_sharing = parameter_sharing
        super(QConv2D, self).__init__(kernel_circuit, data_circuit, operators,
                                      initializer=initializer,
                                      regularizer=regularizer,
                                      constraint=constraint,
                                      trainable=trainable, 
                                      seed=seed,
                                      name=name, **kwargs)
        
    def _validate_init(self):
        if not all(self.kernel_size):
            raise ValueError('The argument `kernel_size` cannot contain 0(s). '
                           'Received: %s' % (self.kernel_size,))

        if not all(self.strides):
            raise ValueError('The argument `strides` cannot contains 0(s). '
                           'Received: %s' % (self.strides,))
        kernel_size = self.kernel_size[0]*self.kernel_size[1]
        num_input_symbols = len(self._base_input_symbols_list)
        if num_input_symbols != kernel_size:
            raise ValueError(f"The kernel size (={kernel_size}) must match the number of "
                             f"input symbols in the data circuit (={num_input_symbols})")
            
    def _validate_build(self, input_shape):
        pass
            
    def build(self, input_shape):
        """Keras build function."""
        assert len(input_shape) in [3, 4]
        if len(input_shape) == 3:
            self.input_rows = input_shape[1]
            self.input_cols = input_shape[2]
            self.input_channels = 1
        else:
            self.input_rows = input_shape[1]
            self.input_cols = input_shape[2]
            self.input_channels = input_shape[3]
        output_shape = get_output_shape(input_shape[1:3], self.kernel_size, self.strides, self.padding)
        self.output_rows = output_shape[0]
        self.output_cols = output_shape[1]
        
        # kernel shape = (filters, input_channels, circuit_params)
        if self.parameter_sharing:
            self.kernel_shape = tf.TensorShape([self.filters, self.input_channels, self._symbols.shape[0]])
        else:
            self.kernel_shape = tf.TensorShape([self.filters, self.input_channels, 
                                               self.output_rows,
                                               self.output_cols,
                                               self._symbols.shape[0]])

        self.kernel = self.add_weight(
                        name='kernel',
                        shape=self.kernel_shape,
                        initializer=self.initializer,
                        regularizer=self.regularizer,
                        constraint=self.constraint,
                        trainable=True,
                        dtype=self.dtype)

        super().build(input_shape)
            
    def _init_weights(self):
        """Initialize trainable weights
        """
        # do not initialize weight before building
        pass
    
    def _get_input_resolver(self):
        base_resolver = super()._get_input_resolver()
        kernel_size = (1, 1) + self.kernel_size + (1,)
        strides = (1, 1) + self.strides + (1,)
        padding = self.padding.upper()
        batchsize = lambda x: tf.gather(tf.shape(x), 0)
        # planes = number of channels
        planes = self.input_channels
        rows = self.input_rows
        cols = self.input_cols
        depth = 1
        reshaped_input_ = lambda x: tf.reshape(x, shape=(batchsize(x), rows, cols, planes))
        # change input order to (batchsize, depth, rows, cols)
        transposed_input = lambda x: tf.transpose(reshaped_input_(x), [0, 3, 1, 2])
        reshaped_input = lambda x: tf.reshape(transposed_input(x), 
                                              shape=(batchsize(x), planes, rows, cols, depth))
        input_patches = lambda x: tf.extract_volume_patches(reshaped_input(x),
            ksizes=kernel_size, strides=strides, padding=padding)
        resolved_input = lambda x: base_resolver(input_patches(x))
        return resolved_input
    
    def call(self, inputs):
        """Keras call function."""
        batchsize = tf.gather(tf.shape(inputs), 0)
        depth = self.input_channels
        rows = self.output_rows
        cols = self.output_cols
        resolved_inputs__ = self._input_resolver(inputs)
        resolved_inputs_ = tf.reshape(resolved_inputs__, [batchsize, depth, 
                                                          self.output_rows, 
                                                          self.output_cols,
                                                          self._num_formulas])
        # change to (depth, batchsize, rows, cols, symbols)
        resolved_inputs = tf.transpose(resolved_inputs_, [1, 0, 2, 3, 4])
        # total number of circuit = filters*depth*batchsize*rows*cols
        circuit_size = tf.reduce_prod([self.filters, batchsize, depth, rows, cols])
        # tile inputs to (filters, depth, batchsize, rows, cols, symbols)
        tiled_up_inputs_ = tf.tile([resolved_inputs], [self.filters, 1, 1, 1, 1, 1])
        # reshape inputs to (circuit_size, symbols)
        tiled_up_inputs = tf.reshape(tiled_up_inputs_, (circuit_size, tf.shape(tiled_up_inputs_)[-1]))
        if self.parameter_sharing:
            # tile size for weights = batchsize*rows*cols
            tile_size = tf.reduce_prod([batchsize, rows, cols])
            tiled_up_weights__ = tf.tile([self.kernel], [tile_size, 1, 1, 1])
            # change to (filters, depth, batchsize*rows*cols, weight_symbols)
            tiled_up_weights_ = tf.transpose(tiled_up_weights__, [1, 2, 0, 3])
        else:
            # tile size for weights = batchsize
            # weight now has shape (batchsize, filters, depth, rows, cols, weight_symbols)
            tiled_up_weights__ = tf.tile([self.kernel], [batchsize, 1, 1, 1, 1, 1])
            # change to (filters, depth, batchsize, rows, cols, weight_symbols)
            tiled_up_weights_ = tf.transpose(tiled_up_weights__, [1, 2, 0, 3, 4, 5])
        # reshape to (circuit_size, weight_symbols)
        tiled_up_weights = tf.reshape(tiled_up_weights_, (circuit_size, tf.shape(tiled_up_weights_)[-1]))
        tiled_up_parameters = tf.concat([tiled_up_inputs, tiled_up_weights], 1)
        
        tiled_up_data_circuit = tf.tile(self._data_circuit, [circuit_size])
        tiled_up_model = tf.tile(self._model_circuit, [circuit_size])
        model_appended = self._append_layer(tiled_up_data_circuit, append=tiled_up_model)
        tiled_up_operators = tf.tile(self._operators, [circuit_size, 1])
        # this is disabled to make autograph compilation easier.
        # pylint: disable=no-else-return
        if self._analytic:
            result = self._executor(model_appended,
                                   symbol_names=self._all_symbols,
                                   symbol_values=tiled_up_parameters,
                                   operators=tiled_up_operators)
        else:
            tiled_up_repetitions = tf.tile(self._repetitions,
                                           [circuit_batch_dim, 1])
            result =  self._executor(model_appended,
                                    symbol_names=self._all_symbols,
                                    symbol_values=tiled_up_parameters,
                                    operators=tiled_up_operators,
                                    repetitions=tiled_up_repetitions)
        averaged_measurement = tf.reduce_mean(result, axis=-1)
        reshaped_output = tf.reshape(averaged_measurement, 
                          (self.filters, self.input_channels, batchsize, self.output_rows, self.output_cols))
        summed_output = tf.reduce_mean(reshaped_output, axis=1)
        final_output = tf.transpose(summed_output, [1, 2, 3, 0])
        return tf.reshape(final_output, (batchsize, self.output_rows, self.output_cols, self.filters))
        # pylint: enable=no-else-return


from math import ceil
import tensorflow as tf

def get_paddings(input_shape, filter_shape, strides):
    """
    Arguments:
        input_shape:
            `(rows, cols)`
        filter_shape:
            `(rows, cols)`
        strides:
            `(rows, cols)`
    """
    assert len(input_shape) == 2
    out_height = ceil(float(input_shape[0]) / float(strides[0]))
    out_width  = ceil(float(input_shape[1]) / float(strides[1]))
    if (input_shape[0] % strides[0] == 0):
        pad_along_height = max(filter_shape[0] - strides[0], 0)
    else:
        pad_along_height = max(filter_shape[0] - (input_shape[0] % strides[0]), 0)
    if (input_shape[1] % strides[1] == 0):
        pad_along_width = max(filter_shape[1] - strides[1], 0)
    else:
        pad_along_width = max(filter_shape[1] - (input_shape[1] % strides[1]), 0)
    pad_top = pad_along_height // 2
    pad_bottom = pad_along_height - pad_top
    pad_left = pad_along_width // 2
    pad_right = pad_along_width - pad_left
    return tf.constant([[pad_top, pad_bottom,], [pad_left, pad_right]])

def get_padded_shape(input_shape, filter_shape, strides, padding="same"):
    """
    Arguments:
        input_shape:
            `(rows, cols)`
        filter_shape:
            `(rows, cols)`
        strides:
            `(rows, cols)`
    """
    assert len(input_shape) == 2
    if padding.lower() == "same":
        paddings = get_paddings(input_shape, filter_shape, strides)
        rows = input_shape[0]+paddings[0][0]+paddings[0][1]
        cols = input_shape[1]+paddings[1][0]+paddings[1][1]
    elif padding.lower() == "valid":
        rows = input_shape[0]
        cols = input_shape[1]
    else:
        raise ValueError("invalid padding: {}".format(padding))
        
    return tf.TensorShape([rows, cols])

# reference: https://cs231n.github.io/convolutional-networks/
def get_output_shape(input_shape, filter_shape, strides, padding="same"):
    """
    Arguments:
        input_shape:
            `(rows, cols)`
        filter_shape:
            `(rows, cols)`
        strides:
            `(rows, cols)`
    """
    assert len(input_shape) == 2
    padded_shape = get_padded_shape(input_shape, filter_shape, strides, padding)
    new_rows = ceil(float(padded_shape[0] - filter_shape[0] + 1) / float(strides[0]))
    new_cols = ceil(float(padded_shape[1] - filter_shape[1] + 1) / float(strides[1]))
    return tf.TensorShape([new_rows, new_cols])


from typing import Sequence, Optional, Union, List, Callable, Tuple
import sympy as sp
import numpy as np

import cirq
import quple
from quple import TemplateCircuitBlock, ParameterisedCircuit

class ParameterisedBlock(TemplateCircuitBlock):
    def __init__(self, copies:int=1,
                       rotation_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                            List[str],List[cirq.Gate],List[Callable],
                                            List['TemplateCircuitBlock']]] =None,
                       entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, 'TemplateCircuitBlock',
                                            List[str],List[cirq.Gate],List[Callable],
                                            List['TemplateCircuitBlock']]] =None,
                       entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                            List[Callable[[int,int],List[Tuple[int]]]]]]=None):
        self.copies = copies
        self.rotation_blocks = rotation_blocks
        self.entanglement_blocks = entanglement_blocks
        self.entangle_strategy = entangle_strategy
        self._num_block_qubits = 2
        
    def build(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int]):

        block = ParameterisedCircuit(circuit.qr.get(qubits), copies=self.copies,
                                     rotation_blocks=self.rotation_blocks,
                                     entanglement_blocks=self.entanglement_blocks,
                                     parameter_index=circuit._parameter_index)
        circuit.append(block)   
    
    @property
    def num_block_qubits(self) -> int:
        return self._num_block_qubits


from typing import Sequence
import numpy as np

import cirq
from cirq import ops


def RXX(theta: float) -> ops.XXPowGate:
    """The XX Ising coupling gate, a native two-qubit operation in ion traps.

    A rotation around the XX axis in the two-qubit bloch sphere.

    The gate implements the following unitary:

        exp(-i θ XX) = [ cos(θ)   0        0       -isin(θ)]
                       [ 0        cos(θ)  -isin(θ)  0      ]
                       [ 0       -isin(θ)  cos(θ)   0      ]
                       [-isin(θ)  0        0        cos(θ) ]

    Args:
        theta: float, sympy.Basic
        	The rotation angle in radians.

    Returns:
        RXX gate with a rotation of `theta` angle.
    """
    return ops.XXPowGate(exponent=theta * 2 / np.pi, global_shift=-0.5)

def RYY(theta: float) -> ops.YYPowGate:
    """The YY Ising coupling gate

    A rotation around the YY axis in the two-qubit bloch sphere.

    The gate implements the following unitary:

        exp(-i θ YY) = [ cos(θ)   0        0       sin(θ)]
                       [ 0        cos(θ)  -isin(θ)  0      ]
                       [ 0       -isin(θ)  cos(θ)   0      ]
                       [sin(θ)  0        0        cos(θ) ]

    Args:
        theta: float, sympy.Basic
        	The rotation angle in radians.

    Returns:
        RYY gate with a rotation of `theta` angle.
    """
    return ops.YYPowGate(exponent=theta * 2 / np.pi, global_shift=-0.5)

def RZZ(theta: float) -> ops.ZZPowGate:
    """The ZZ Ising coupling gate

    A rotation around the ZZ axis in the two-qubit bloch sphere.

    The gate implements the following unitary:

        exp(-i θ ZZ) = [ exp(iθ/2)     0        0           0     ]
                       [    0      exp(-iθ/2)   0           0     ]
                       [    0          0     exp(-iθ/2)     0     ]
                       [    0          0        0       exp(iθ/2) ]

    Args:
        rads: float, sympy.Basic
        	The rotation angle in radians.

    Returns:
        RZZ gate with a rotation of `theta` angle.
    """
    return ops.ZZPowGate(exponent=theta * 2 / np.pi, global_shift=-0.5)    

class CompositeGate:
    pass


class PauliRotation(CompositeGate):
    def __init__(self, pauli_string, theta, global_shift=False):
        self.paulis = pauli_string[::-1]
        self.theta = theta
        self.indices = [i for i, pauli in enumerate(self.paulis) if pauli != 'I']
        self.global_shift = global_shift

    @staticmethod
    def change_basis(*qubits:Sequence[int],
                     pauli_string:Sequence[str], inverse=False) -> None:
        # do not change basis if only first order pauli operator
        if len(pauli_string) == 1:
            return
        operations = []
        for i, pauli in enumerate(pauli_string):
            if pauli == 'X':
                operations.append(ops.H(qubits[i]))
            elif pauli == 'Y':
                if inverse:
                    operations.append(ops.rx(-np.pi / 2)(qubits[i]))
                else: 
                    operations.append(ops.rx(np.pi / 2)(qubits[i]))
        return operations 

    def __call__(self, *qubits):
        if not self.indices:
            return None

        operations = []

        operations += PauliRotation.change_basis(*qubits, pauli_string=self.paulis)

        qpairs = [(qubits[i], qubits[i+1]) for i in range(len(qubits)-1)]

        for qpair in qpairs:
            operations.append(cirq.CX(*qpair))
        
        # do not switch to RZ gate if only first order pauli operator
        if len(self.paulis) == 1:
            if self.global_shift:
                if self.paulis[0] == 'Z':
                    operations.append(ops.ZPowGate(self.theta)(qubits[-1]))
                elif self.paulis[0] == 'X':
                    operations.append(ops.XPowGate(self.theta)(qubits[-1]))
                elif self.paulis[0] == 'Y':
                    operations.append(ops.YPowGate(self.theta)(qubits[-1]))             
            else:
                if self.paulis[0] == 'Z':
                    operations.append(ops.rz(self.theta)(qubits[-1]))
                elif self.paulis[0] == 'X':
                    operations.append(ops.rx(self.theta)(qubits[-1]))
                elif self.paulis[0] == 'Y':
                    operations.append(ops.ry(self.theta)(qubits[-1]))    
        else:
            if self.global_shift:
                operations.append(ops.ZPowGate(self.theta)(qubits[-1]))
            else:
                operations.append(ops.rz(self.theta)(qubits[-1]))

        for qpair in qpairs[::-1]:
            operations.append(cirq.CX(*qpair))
        
        operations += PauliRotation.change_basis(*qubits, pauli_string=self.paulis, inverse=True)
        
        return operations


from typing import Callable, Optional
import numpy as np
import sympy as sp

from quple.data_encoding.general_pauli_z_encoding import GeneralPauliZEncoding

class FirstOrderPauliZEncoding(GeneralPauliZEncoding):
    """The first order Pauli Z encoding circuit
    
    An encoding circuit consisting of layers of unitary operators of the
    form exp(iψ(x)Z)H^{⊗n} where ψ is a data encoding function, Z is the 
    Pauli Z operator and and x = (x_1, . . . , x_n ) are the input features
    to be encoded.
    
    To encode data of feature dimension n, a set of general Pauli operators 
    are chosen to encode the data into an n qubit circuit. Each Pauli operator
    will contribute to a unitary operation \exp(i\sum_{s\in S}ψ_s(x_s)Z_s)H^{⊗n}
    where S is the set of qubit indices for all qubits in the circuit.

    The Pauli encoding uses the circuit-centric design with alternating
    rotation and entanglement layer to achieve a strongly entangling circuit. 
    Ref: https://arxiv.org/pdf/1804.00633.pdf

    Examples:
    >> cq = FirstOrderPauliZEncoding(feature_dimension=4, copies=2)
    >> cq
    (0, 0): ───H───Rz(pi*x_0)───H───Rz(pi*x_0)───

    (0, 1): ───H───Rz(pi*x_1)───H───Rz(pi*x_1)───

    (0, 2): ───H───Rz(pi*x_2)───H───Rz(pi*x_2)───

    (0, 3): ───H───Rz(pi*x_3)───H───Rz(pi*x_3)───
    
    Args:
        feature_dimension: int
            Dimension of data to be encoded (=number of qubits in the circuit)
        copies: int
            Number of repetition of the encoding circuit
        encoding_map: function that maps a numpy array to a number
            Data mapping function from R^(feature_dimension) to R
        name: str
            Name of the encoding circuit
    """
    def __init__(self, feature_dimension: int,
                 copies:int=2, 
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 global_shift:bool=False,    
                 parameter_symbol:str='x',
                 parameter_scale=sp.pi,
                 name:str='FirstOrderPauliZEncoding', *args, **kwargs):
        '''Creates first order Pauli Z encoding circuit

        
        Args:
            feature_dimension: int
                Dimension of data to be encoded (=number of qubits in the circuit)
            copies: int
                Number of repetition of the encoding circuit
            encoding_map: function that maps a numpy array to a number
                Data mapping function from R^(feature_dimension) to R
            global_shift: bool
                Include global shift to RZ gate                
            name: str
                Name of the encoding circuit
        '''        
        super().__init__(feature_dimension, copies=copies, z_order=1,
                         encoding_map=encoding_map, global_shift=global_shift,
                         parameter_symbol=parameter_symbol,
                         parameter_scale=parameter_scale,
                         name=name, *args, **kwargs)


import numbers
import numpy as np
import sympy
from sympy.core import numbers as sympy_numbers
from sympy.functions.elementary.trigonometric import TrigonometricFunction, InverseTrigonometricFunction

import tensorflow as tf

tf_ops_map = {
    sympy.sin: tf.sin,
    sympy.cos: tf.cos,
    sympy.tan: tf.tan,
    sympy.asin: tf.asin,
    sympy.acos: tf.acos,
    sympy.atan: tf.atan,
    sympy.atan2: tf.atan2,
    sympy.cosh: tf.cosh,
    sympy.tanh: tf.tanh,
    sympy.sinh: tf.sinh
}

def stack(func, lambda_set, intermediate=None):
    if intermediate is None:
        return stack(func, lambda_set[1:], lambda_set[0])
    if len(lambda_set) > 0:
        new_lambda = lambda x:func(intermediate(x), lambda_set[0](x))
        return stack(func, lambda_set[1:], new_lambda)
    else:
        return intermediate
    
def resolve_value(val):
    if isinstance(val, numbers.Number) and not isinstance(val, sympy.Basic):
        return tf.constant(float(val), dtype=tf.float32)
    elif isinstance(val, (sympy_numbers.IntegerConstant, sympy_numbers.Integer)):
        return tf.constant(float(val.p), dtype=tf.float32)
    elif isinstance(val, (sympy_numbers.RationalConstant, sympy_numbers.Rational)):
        return tf.divide(tf.constant(val.p, dtype=tf.float32), tf.constant(val.q, dtype=tf.float32))
    elif val == sympy.pi:
        return tf.constant(np.pi, dtype=tf.float32)
    else:
        return NotImplemented    

def resolve_formulas(formulas, symbols):
    lambda_set = [resolve_formula(f, symbols) for f in formulas]
    stacked_ops = stack(lambda x, y:tf.concat((x, y), 0), lambda_set)
    n_formula = tf.constant([len(formulas)])
    transposed_x = lambda x: tf.transpose(x, perm=tf.roll(tf.range(tf.rank(x)), shift=1, axis=0))
    resolved_x = lambda x: stacked_ops(transposed_x(x))
    reshaped_x = lambda x: tf.reshape(resolved_x(x), 
                           tf.concat((n_formula, tf.strided_slice(tf.shape(x), begin=[0], end=[-1])), axis=0))
    transformed_x = lambda x: tf.transpose(reshaped_x(x), perm=tf.roll(tf.range(tf.rank(x)), shift=-1, axis=0))
    return transformed_x
           
def resolve_formulas_legacy(formulas, symbols):
    lambda_set = [resolve_formula(f, symbols) for f in formulas]
    stacked_ops = stack(lambda x, y:tf.concat((x, y), 0), lambda_set)
    n_formula = len(formulas)
    transformed_x = lambda x: tf.transpose(tf.reshape(stacked_ops(tf.transpose(x)), (n_formula, tf.gather(tf.shape(x), 0))))
    return transformed_x                               
    
def resolve_formula(formula, symbols):
    
    # Input is a pass through type, no resolution needed: return early
    value = resolve_value(formula)
    if value is not NotImplemented:
        return lambda x:value
    
    # Handles 2 cases:
    # formula is a string and maps to a number in the dictionary
    # formula is a symbol and maps to a number in the dictionary
    # in both cases, return it directly.
    if formula in symbols:
        index = symbols.index(formula)
        return lambda x:x[index]
                               
    # formula is a symbol (sympy.Symbol('a')) and its string maps to a number
    # in the dictionary ({'a': 1.0}).  Return it.
    if isinstance(formula, sympy.Symbol) and formula.name in symbols:
        index = symbols.index(formula.name)
        return lambda x:x[index]
    
    # the following resolves common sympy expressions
    if isinstance(formula, sympy.Add):
        addents = [resolve_formula(arg, symbols) for arg in formula.args]
        return stack(tf.add, addents)
    
    if isinstance(formula, sympy.Mul):
        factors = [resolve_formula(arg, symbols) for arg in formula.args]
        return stack(tf.multiply, factors)
    
    if isinstance(formula, sympy.Pow) and len(formula.args) == 2:
        base = resolve_formula(formula.args[0], symbols)
        exponent = resolve_formula(formula.args[1], symbols)
        return lambda x: tf.pow(base(x), exponent(x))
    
    if isinstance(formula, sympy.Pow):
        base = resolve_formula(formula.args[0], symbols)
        exponent = resolve_formula(formula.args[1], symbols)
        return lambda x: tf.pow(base(x), exponent(x))
    
    if isinstance(formula, (TrigonometricFunction, InverseTrigonometricFunction)):
        ops = tf_ops_map.get(type(formula), None)
        if ops is None:
            raise ValueError("unsupported sympy operation: {}".format(type(formula)))
        arg = resolve_formula(formula.args[0], symbols)
        return lambda x: ops(arg(x))
    



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSIONFILE="quple/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))        

setuptools.setup(
    name="quple", # Replace with your own username
    version=verstr,
    author="Alkaid Cheng",
    author_email="chi.lung.cheng@cern.ch",
    description="A framework for quantum machine learning in high energy physics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cern.ch/clcheng/quple",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[   
          'cirq',
          'numpy',
          'imageio',
          'tensorflow_quantum'
      ],
    python_requires='>=3.5',
)



import numpy as np
import numba
from sklearn import svm
import quple
from sklearn.model_selection import GridSearchCV
from .qsvm_logger import QSVMLogger

@numba.vectorize([numba.float64(numba.complex128),numba.float32(numba.complex64)])
def abs2(x):
    return x.real**2 + x.imag**2


class QSVM:
    def __init__(self):
        pass
    
    @staticmethod
    def get_kernel_matrix(state_vectors_left, state_vectors_right):
        return abs2(state_vectors_left.conjugate() @ state_vectors_right.T)
    
    @staticmethod
    def run(encoding_circuit, x_train, y_train, x_test, y_test, 
            random_seed = 0, backend=None, **kwargs):
        from qiskit.aqua.algorithms import QSVM as QSVM_qiskit
        attributes = {}
        attributes['encoding_circuit'] = type(encoding_circuit).__name__
        attributes['feature_dimension'] = x_train.shape[1]
        attributes['train_size'] = x_train.shape[0]
        attributes['test_size'] = x_test.shape[0]
        attributes['n_qubit'] = encoding_circuit.num_qubits
        attributes['encoding_map'] = encoding_circuit._data_map_func.__name__ 
        logger = QSVMLogger(attributes)      
        logger.info('Evaluating kernel matrix for train data...')
        train_kernel_matrix = QSVM_qiskit.get_kernel_matrix(backend, encoding_circuit, x_train)
        logger.info('Evaluating kernel matrix for test data...')        
        train_test_kernel_matrix = QSVM_qiskit.get_kernel_matrix(backend, encoding_circuit, x_test, x_train)
        logger.info('Training started')
        svc = svm.SVC(kernel='precomputed', probability=True, random_state=random_seed)

        tune_params = {
                       'C': [1,10,100,200,400,800],
                       'gamma': [0.01,0.1,1]
                      }

        clf = GridSearchCV(estimator=svc, param_grid=tune_params, n_jobs=-1, cv=3, scoring='roc_auc')

        clf.fit(train_kernel_matrix,y_train)

        logger.on_train_end(clf, train_kernel_matrix, train_test_kernel_matrix, y_test)
        return clf


import cirq
from quple.components.gate_ops import RXX, RYY, RZZ

class CirqInterface(ABC):
    __GATE_MAPPING__ = {
        "H": cirq.H, # Hadamard gate
        "I": cirq.I,  # one-qubit Identity gate
        "S": cirq.S, # Clifford S gate
        "T": cirq.T, # non-Clifford T gate
        'X': cirq.X, # Pauli-X gate
        "Y": cirq.Y, # Pauli-Y gate
        "Z": cirq.Z, # Pauli-Z gate
        "PauliX": cirq.X, # Pauli-X gate
        "PauliY": cirq.Y, # Pauli-Y gate
        "PauliZ": cirq.Z, # Pauli-Z gate
        "CX": cirq.CX, # Controlled-NOT gate
        "CNOT": cirq.CNOT, # Controlled-NOT gate
        "CZ": cirq.CZ, # Controlled-Z gate
        "XX": cirq.XX, # tensor product of two X gates (X parity gate)
        "YY": cirq.YY, # tensor product of two Y gates (Y parity gate)
        "ZZ": cirq.ZZ, # tensor product of two Z gates (Z parity gate)
        "XPowGate": cirq.XPowGate, # rotation along X axis with extra phase factor
        "YPowGate": cirq.YPowGate, # rotation along Y axis with extra phase factor 
        "ZPowGate": cirq.ZPowGate, # rotation along Z axis with extra phase factor
        "XXPowGate": cirq.XXPowGate, # X parity gate raised to some power
        "YYPowGate": cirq.YYPowGate, # Y parity gate raised to some power
        "ZZPowGate": cirq.ZZPowGate, # Z parity gate raised to some power
        "MS": cirq.ms, # Mølmer–Sørensen gate == RXX (A rotation around the XX axis in the two-qubit bloch sphere)
        "RXX": RXX, # XX Ising coupling gate (A rotation around the XX axis in the two-qubit bloch sphere)
        "RXX": RYY, # YY Ising coupling gate (A rotation around the YY axis in the two-qubit bloch sphere)
        "RXX": RZZ, # ZZ Ising coupling gate (A rotation around the ZZ axis in the two-qubit bloch sphere)                        
        "RX": cirq.rx, # rotation along X axis
        "RY": cirq.ry, # rotation along Y axis
        "RZ": cirq.rz, # rotation along Z axis
        "CCNOT": cirq.CCNOT, # Toffoli gate
        "CCX": cirq.CCX, # Toffoli gate
        "Toffoli": cirq.TOFFOLI, # Toffoli gate
        "SWAP": cirq.SWAP, # SWAP gate
        "CSWAP": cirq.CSWAP, # Controlled SWAP gate
        "ISWAP": cirq.ISWAP, # ISWAP gate
        "RISWAP": cirq.riswap, #Rotation ISWAP gate (X⊗X + Y⊗Y)
        "FSim": cirq.FSimGate, # Fermionic simulation gate
        "Fredkin": cirq.FREDKIN, # Controlled SWAP gate
        "CXPowGate": cirq.CXPowGate, # Controlled Power of an X gate
        "CZPowGate": cirq.CZPowGate, # Controlled Power of an Z gate
        "CNOTPowGate": cirq.CXPowGate, # Controlled Power of an X gate
    }


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit

class ExcitationPreserving(ParameterisedCircuit):
    '''The excitation preserving parameterised circuit
    
    Implementation based on the Qiskit library.
    
    The ExcitationPreserving circuit consists of layers 
    of the two qubit Fermionic simulation, or fSim, gate set. 
    Under this gate set, the σ_Xσ_X and σ_Yσ_Y couplings 
    between the qubits have equal coefficients which
    conserves the number of excitations of the qubits. 
    Algorithms performed with just Pauli Z rotations 
    and fSim gates enable error mitigation techiques
    including post selection and zero noise extrapolation.
    Referemce: https://arxiv.org/pdf/1805.04492.pdf
    
    Further reference: https://arxiv.org/pdf/2001.08343.pdf
    
    The ExcitationPreserving circuit may be used as a variational
    circuit or a model circuit in the PQC layer of a machine 
    learning model. It represents a unitary U(θ) parameterised 
    by a set of free  parameters θ. The circuit may be repeated
    several times to  increase the number of parameters and 
    hence the complexity of the model. 
    
    It is an example of a strongly entangling circuit which has 
    the advantage of capturing correlations between the data
    features at all ranges for a short range circuit. 
    Reference: https://arxiv.org/pdf/1804.00633.pdf
    
    Args:
      n_qubit: int
          Number of qubits in the circuit
      copies: int
         Number of times the layers are repeated.
      entanglement_gate: 'RISWAP' or 'FSim', default='RISWAP'
         Excitation preserving gate operation to use for entangling qubits
      entangle_strategy: Determines how the qubits are connected in an entanglement block.
         If str, it specifies the name of the strategy.
         If callable, it specifies the function to map to an interaction graph.
         If list of str, it specifies the names of a list of strategies. The 
         strategy to use is decided by the current block index. For example, if
         the circuit is building the n-th entanglement block in the entanglement layer,
         then the n-th strategy in the list will be used.
         If list of callable, it specifies the list of functions to map to an interaction
         graph. The function to use is decided by the current block index.
      parameter_symbol: str
         Symbol prefix for circuit parameters.
      name: str
         Name of the circuit.
    '''        
    def __init__(self, n_qubit: int, copies: int=2, 
                 entanglement_gate:str='RISWAP',
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str='ExcitationPreserving', *args, **kwargs):
        
        allowed_blocks = ['RISWAP', 'FSim']
        if entanglement_gate not in allowed_blocks:
            raise ValueError('Unsupported gate operation {}, choose one of '
                             '{}'.format(entanglement_gate, allowed_blocks))
 
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks='RZ',
                         entanglement_blocks=entanglement_gate,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         name=name,
                         final_rotation_layer=True,
                         flatten_circuit=False,
                         *args, **kwargs)


from enum import Enum
import numpy as np
import numba

class DataPrecision(Enum):
    single = 1
    double = 2

@numba.vectorize([numba.float64(numba.complex128),numba.float32(numba.complex64)])
def abs2(x):
    # element-wise modulus square
    return x.real**2 + x.imag**2

def gramian_matrix(A:np.ndarray, B:np.ndarray):
    # The Gramian matrix
    # G = AB*
    return A @ B.conjugate().T

def split_gramian_matrix(A:np.ndarray, B:np.ndarray, n_split:int=1):
    if n_split == 1:
        return gramian_matrix(A, B)
    # output dimension
    dimension = A.shape[0]
    if A.shape != B.shape:
        raise ValueError('Input matrices must have the same shape')
    if A.dtype != B.dtype:
        raise ValueError('Input matrices must have the same data type')
    if dimension % n_split != 0:
        raise ValueError('Dimension of input matrix must divide the number of splits')
    # size of each split
    k = dimension//n_split 
    G = np.zeros((dimension, dimension), dtype=A.dtype)
    for i in range(n_split):
        for j in range(n_split):
            G[i*k:(i+1)*k, j*k:(j+1)*k] = gramian_matrix(A[i*k:(i+1)*k], B[j*k:(j+1)*k])
    return G


from typing import Optional, Dict

import matplotlib.pyplot as plt

def visualize_images(images:"numpy.ndarray", columns:int=8, cmap='viridis',
                     subplots_options:Optional[Dict]=None, axis_off:bool=False, colorbar:bool=True,
                     labels:Optional["numpy.ndarray"]=None, label_map:Optional[Dict]=None,
                     save_as:str=None):
    """Visualize a set of 2D images in a grid
    
    Arguments:
        images: numpy.ndarray
            A 3D numpy array representing batches of images of shape (batchsize, rows, cols).
        colunms: int
            Number of coumns in the grid.
        subplots_options: (Optional) dict
            A dictionary containing subplots options.
        labels: (Optional) numpy.ndarray
            A 1D numpy array representing the object labels to the images of shape (batchsize).
        label_map: (Optional) dict
            A dictionary containing the map from an object label to a string representation of
            that label which will be displayed in the image title. If None, images are displayed
            without titles.
        save_as: (Optional) str
            The path from which output image is saved. If None, the output image will not be saved.
    Returns:
        matplotlib.pyplot object
    """
    plt.clf()
    size = images.shape[0]
    rows = ( size // columns ) + 1
    fig = plt.figure(figsize=(20, rows*3))
    if (labels is not None) and (label_map is not None):
        assert labels.shape[0] == images.shape[0]
        titles = ["Image {}: {}".format(i, label_map[labels[i]]) for i in range(size)]
    else:
        titles = [None]*size
    for i in range(images.shape[0]):
        ax = fig.add_subplot(rows, columns, i+1)
        ax.set_title(titles[i])
        im = ax.imshow(images[i], cmap=cmap)
        if axis_off:
            plt.axis('off')
    if colorbar:
        plt.colorbar(im)
    if subplots_options is not None:
        plt.subplots_adjust(**subplots_options)
    if save_as is not None:
        os.makedirs(os.path.dirname(save_as), exist_ok=True)
        plt.savefig(save_as, bbox_inches="tight")
    return plt